"""Functions and mappings for coordinate grid"""
from functools import lru_cache
from typing import Dict

import numpy as np

base_coordinate_mappings: Dict[str, Dict[str, str]] = {
    "cartesian": {"x": "x1", "y": "x2", "z": "x3"},
    "polar": {"r": "x1", "phi": "x2", "z": "x3"},
    "cylindrical": {"r": "x1", "z": "x2", "x3": "x3"},
    "spherical": {"r": "x1", "theta": "x2", "phi": "x3"},
}


@lru_cache()
def mapping_grid(coordinates: str) -> Dict[str, str]:
    """Generate variable name mapping for specified coordinate system.

    Implements mapping for all coordinates (cell edges and centers) as well as
    velocity components.

    Args:
        coordinates (str): coordinate system name (cartesian, polar, cylindrical, spherical)

    Returns:
        dict[str, str]: Mapping from coordinate system dependend name to PLUTO name
    """
    if coordinates not in base_coordinate_mappings:
        raise NotImplementedError(f"Coordinate system {coordinates} not implemented")

    mapping = base_coordinate_mappings[coordinates].copy()
    grid_mappings = {}
    for coord_name, coord_num in mapping.items():
        grid_mappings[f"{coord_name}i"] = f"{coord_num}i"
        grid_mappings[f"d{coord_name}"] = f"d{coord_num}"
        grid_mappings[f"L{coord_name}"] = f"Lx{coord_num}"
    mapping.update(grid_mappings)
    return mapping


@lru_cache()
def mapping_vars(coordinates: str) -> Dict[str, str]:
    """Coordinate name mapping for velocity components

    Note:
        Names for magnetic and radiative variables are always included.

    Args:
        coordinates (str): coordinate system name (cartesian, polar, cylindrical, spherical)

    Returns:
        dict[str, str]: Mapping from coordinate system dependend name to PLUTO name
    """
    if coordinates not in base_coordinate_mappings:
        raise NotImplementedError(f"Coordinate system {coordinates} not implemented")

    mapping = {}
    for coord_name, coord_num in base_coordinate_mappings[coordinates].items():
        # velocity components
        mapping[f"v{coord_name}"] = f"v{coord_num}"
        # magnetic field
        mapping[f"B{coord_name}"] = f"B{coord_num}"
        mapping[f"B{coord_name}s"] = f"B{coord_num}s"
        # radiativ flux
        mapping[f"fr{coord_name}"] = f"fr{coord_num[1:]}"

    return mapping


tex_chars = {
    "theta": r"\theta",
    "rho": r"\rho",
    "phi": r"\phi",
    "prs": "P",
    "enr": "E_r",
    "Bs": r"B^{(s)}",
    "fr": r"F^{(r)}",
}


@lru_cache()
def mapping_tex(coordinates: str) -> Dict[str, str]:
    """Coordinate and variable mapping to LaTeX math mode symbols.

    This maps from variables/coordinates as named in PLUTO outputs
    to Latex names in the respective coordinate system.
    This is useful for plotting.

    Args:
        coordinates (str): coordinate system name (cartesian, polar, cylindrical, spherical)

    Returns:
        dict[str, str]: Mapping from PLUTO name to LaTeX name in coordinate system
    """
    if coordinates not in base_coordinate_mappings:
        raise NotImplementedError(f"Tex mappings for {coordinates} not implemented")

    mapping = {}
    for coord_name, coord_num in base_coordinate_mappings[coordinates].items():
        mapping[coord_num] = mapping[coord_name] = tex_chars.get(coord_name, coord_name)

    mapping_vars = {}
    for coord, tex in mapping.items():
        # velocity
        mapping_vars[f"v{coord}"] = f"v_{tex}"
        # magnetic field
        mapping_vars[f"B{coord}"] = f"B_{tex}"
        mapping_vars[f"B{coord}s"] = f"{tex_chars['Bs']}_{tex}"
        # radiative flux
        mapping_vars[f"fr{coord}"] = f"{tex_chars['fr']}_{tex}"
    # fix radiative flux naming
    for i in range(1, 4):
        try:
            mapping_vars[f"fr{i}"] = mapping_vars.pop(f"frx{i}")
        except KeyError:
            pass
    mapping.update(mapping_vars)

    for key in ("rho", "prs"):
        mapping[key] = tex_chars.get(key, key)

    return mapping


def transform_mesh(grid, mesh1, mesh2):
    if grid.coordinates == "cartesian":
        return (
            f"{grid.mapping_tex[f'x{grid.rdims_ind[0]+1}']}",
            f"{grid.mapping_tex[f'x{grid.rdims_ind[1]+1}']}",
        ), (mesh1, mesh2)
    elif grid.coordinates == "spherical":
        if grid.rdims_ind == (0, 1):
            r = mesh1
            z = r * np.cos(mesh2)
            return ("r", "z"), (r, z)
        elif grid.rdims_ind == (0, 2):
            factor = mesh1 * np.sin(grid.x2[0])
            x = factor * np.cos(mesh2)
            y = factor * np.sin(mesh2)
            return ("x", "y"), (x, y)
        raise NotImplementedError("Projection in (theta, phi) not supported")
    elif grid.coordinates == "cylindrical":
        return ("r", "z"), (mesh1, mesh2)
    elif grid.coordinates == "polar":
        if grid.rdims_ind == (0, 1):
            x = mesh1 * np.cos(mesh2)
            y = mesh1 * np.sin(mesh2)
            return ("x", "y"), (x, y)
        elif grid.rdims_ind == (0, 2):
            return ("r", "z"), (mesh1, mesh2)
        else:
            raise NotImplementedError("Projection in (phi, z) not supported")



import argparse
from pathlib import Path

from . import __version__
from .simulation import Simulation


def main():
    parser = argparse.ArgumentParser(description="Show information on PLUTO simulation")
    parser.add_argument(
        "paths", help="Path to PLUTO simulation", default=["."], nargs="*"
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s {}".format(__version__)
    )

    args = parser.parse_args()

    for path in args.paths:
        print(info(path))


def info(simulationpath):
    """Print info on simulation"""
    sims = []
    for format_ in Simulation.supported_formats:
        try:
            sims.append(Simulation(simulationpath, format=format_))
        except FileNotFoundError:
            pass
    if not sims:
        return "No simulation found at '{}'".format(simulationpath)

    sim = sims[0]
    output = (
        f"PLUTO simulation at '{sim.path}'\n"
        f"{sim.grid.coordinates.capitalize()} grid with dimensions {sim.dims}\n"
        f"Domain: x1: {sim.x1i[0]:.2e} .. {sim.x1i[-1]:.2e} (Lx1 = {sim.Lx1:.2e})\n"
        f"        x2: {sim.x2i[0]:.2e} .. {sim.x2i[-1]:.2e} (Lx2 = {sim.Lx2:.2e})\n"
        f"        x2: {sim.x3i[0]:.2e} .. {sim.x3i[-1]:.2e} (Lx3 = {sim.Lx3:.2e})\n"
        f"Available variables: {' '.join(sim.vars)}\n"
        "Data files:\n"
    )
    for sim in sims:
        output += (
            f"    Format {sim.format}: {len(sim)} files, "
            f"last time {sim.t[-1]}, data timestep {sim.dt.mean():.2e}\n"
        )

    return output


if __name__ == "__main__":
    main()



"""Miscellaneous tools"""


def cached_property(func):
    """Cache class property decorator

    Caches attribute under `_attr_`.
    """

    def wrapper(self):
        cached_name = "_" + func.__name__
        try:
            return getattr(self, cached_name)
        except AttributeError:
            setattr(self, cached_name, func(self))
        return getattr(self, cached_name)

    return property(wrapper)


class Slicer:
    """Helper class to create sliced subclass

    Takes a class to create and arbitrary keyword arguments to pass to the class.
    It then initializes the object when accessed with `slicer[:]`, and passes the
    slice as additional kwarg (with name `slice_`) to the class.
    """

    def __init__(self, SliceClass, **kwargs):
        self.SliceClass = SliceClass
        self.kwargs = kwargs

    def __getitem__(self, slice_):
        return self.SliceClass(slice_=slice_, **self.kwargs)



try:
    import importlib.metadata

    __version__ = importlib.metadata.version("plutoplot")
    del importlib
except ImportError:  # Python <3.8
    import pkg_resources

    __version__ = pkg_resources.get_distribution("plutoplot").version
    del pkg_resources


from . import misc
from .grid import Grid
from .metadata import Definitions_h, Pluto_ini
from .plutodata import PlutoData
from .simulation import Simulation



from pathlib import Path
from typing import Dict

import numpy as np

from plutoplot.misc import Slicer, cached_property

from .coordinates import mapping_grid, mapping_tex, mapping_vars, transform_mesh


class Grid:
    """Grid datastructure to be initialized from gridfile

    Attributes:
        gridfile_path (Path): Path to gridfile
        coordinates (str): Name of coordinate system
        mapping_grid (:obj:`dict` of :obj:`str`): mapping from coordinate system variable
            name to PLUTO variable names.
            (e.g. for spherical coordinates `r`->`x1`, `theta`->`x2`, `phi`->`x3`, )
        mapping_vars (:obj:`dict` of :obj:`str`): mapping from coordinate system variable
            attribute name to PLUTO variable names.
            (e.g. for spherical coordinates `vr`->`vx1`, `vtheta`->`vx2`, `vphi`->`vx3`)
        mapping_tex (:obj:`dict` of :obj:`str`): mapping from variable name to LaTeX names.

        dims (:obj:`tuple` of :obj:`int`): domain dimensions
        size (int): total size of data arrays (product of dims)

        x1, x2, x3 (numpy.ndarray): cell centered grid (1d, not as mesh)
        x1i, x2i, x3i (numpy.ndarray): cell interface coordinates (1d, not as mesh)
        dx1, dx2, dx3 (numpy.ndarray): cell sizes (1d, not as mesh)

        r, z (numpy.ndarray): available if `coordinates == 'cylindrical'`, maps to x1, x2
        ri, zi (numpy.ndarray): available if `coordinates == 'cylindrical'`, maps to x1i, x2i
        dr, dz (numpy.ndarray): available if `coordinates == 'cylindrical'`, maps to dx1, dx2

        r, phi, z (numpy.ndarray): available if `coordinates == 'polar'`, maps to x1, x2, x3
        ri, phii, zi (numpy.ndarray): available if `coordinates == 'polar'`, maps to x1i, x2i, x3i
        dr, dphi, dz (numpy.ndarray): available if `coordinates == 'polar'`, maps to dx1, dx2, dx3

        r, theta, phi (numpy.ndarray): available if `coordinates == 'spherical'`, maps to x1, x2, x3
        ri, thetai, phii (numpy.ndarray): available if `coordinates == 'spherical'`, maps to x1i, x2i, x3i
        dr, dtheta, dphi (numpy.ndarray): available if `coordinates == 'spherical'`, maps to dx1, dx2, dx3

    Todo:
        * Generalize and document meshgrid functions
    """

    def __init__(self, gridfile: Path, coordinates: str = None, indexing="ijk"):
        """Initialize Grid from gridfile

        Args:
            gridfile (:obj:`str` or :obj:`Pathlike`): path to gridfile
            coordinates (:obj:`str`, optional): name of coordinate system (cartesian, polar,
                cylindrical, spherical). If not set this will be read from gridfile.
            indexing (:obj:`str`, optional): index order for arrays. 'ijk' or 'kji'
        """
        self.gridfile_path: Path = Path(gridfile)
        self.coordinates: str = None
        self.mapping_grid: Dict[str, str] = None
        self.mapping_vars: Dict[str, str] = None
        self.mapping_tex: Dict[str, str] = None

        # helper function to transpose arrays if necessary
        if indexing == "ijk":
            self.T = lambda x: x.T  # transpose
        elif indexing == "kji":
            self.T = lambda x: x  # do nothing
        else:
            raise RuntimeError(f"Pluto Grid: indexing {indexing} not supported")
        self.indexing = indexing

        # read gridfile, get coordinate system if necessary
        self.read_gridfile(gridfile, coordinates)

        if coordinates is not None:
            self.set_coordinate_system(coordinates)

        self.slicer = Slicer(GridSlice, grid=self)
        self.slice = slice(None)

    def set_coordinate_system(self, coordinates: str) -> None:
        """Set coordinate system of grid and get name mapping

        Args:
            coordinates (str): name of coordinate system (cartesian, polar,
                                                          cylindrical, spherical)
        """
        self.coordinates = coordinates
        self.mapping_grid = mapping_grid(coordinates)
        self.mapping_vars = mapping_vars(coordinates)
        self.mapping_tex = mapping_tex(coordinates)

    def read_gridfile(self, gridfile_path: Path, coordinates: str = None) -> None:
        """Read and parse gridfile

        Args:
            gridfile_path (:obj:`str` or :obj:`Pathlike`): Path to PLUTO gridfile `grid.out`
            coordinates (:obj:`str`, optional): coordinate system name
                                                If not set this will be read from gridfile

        Sets Attributes:
            x1, x2, x3 (numpy.ndarray): cell centered grid (1d, not as mesh)
            x1i, x2i, x3i (numpy.ndarray): cell interfaces (1d, not as mesh)
            dx1, dx2, dx3 (numpy.ndarray): cell sizes (1d, not as mesh)
            Lx1, Lx2, Lx3 (numpy.ndarray): Domain width
            dims (:obj:`tuple` of :obj:`int`): domain dimensions
            data_shape (:obj:`tuple` of :obj:`int`): shape of data array.
                                                     Depends on index order
            size (int): total size of data arrays (product of dims)
        """
        # to be filled with left and right cell interfaces
        x = []
        dims = []
        with gridfile_path.open() as gf:
            # Gridfile header
            header = False  # marker if gf pointer is in header
            while True:
                line = gf.readline()
                if line.startswith("# *****"):
                    # header starts and ends with # *****...
                    # toggle marker when entering header
                    # and exit when header is finished
                    header = not header
                    if not header:
                        break
                # set coordinate system from gridfile if not explicitly set
                elif coordinates is None and line.startswith("# GEOMETRY"):
                    self.set_coordinate_system(line[11:].strip().lower())

            # read all dimensions
            while True:
                # read line by line, stop if EOF
                line = gf.readline()
                if not line:
                    break
                # find line with resolution in dimension
                splitted = line.split()
                if len(splitted) == 1:
                    dim = int(splitted[0])
                    dims.append(dim)
                    # read all data from dimension, moves file pointer
                    data = np.fromfile(gf, sep=" ", count=dim * 3).reshape(-1, 3)
                    # save left and right cell interface
                    x.append((data[:, 1], data[:, 2]))

        # cell centers
        self.xn = tuple((xn[0] + xn[1]) / 2 for xn in x)
        # cell interfaces
        self.xni = tuple(np.append(xn[0], xn[1][-1]) for xn in x)
        # cell widths
        self.dxn = tuple(x[1:] - x[:-1] for x in self.xni)
        # domain widths
        self.L = tuple(x[-1] - x[0] for x in self.xni)

        # reference in named attributes
        for i in range(3):
            setattr(self, f"x{i+1}", self.xn[i])
            setattr(self, f"x{i+1}i", self.xni[i])
            setattr(self, f"dx{i+1}", self.dxn[i])
            setattr(self, f"Lx{i+1}", self.L[i])

        self.dims = tuple(dims)
        # indices of dims which are not 1
        self.rdims = tuple(dim for dim in dims if dim > 1)
        self.rdims_ind = tuple(i for i, dim in enumerate(dims) if dim > 1)

        self.data_shape = tuple(reversed(dims))
        if self.indexing == "ijk":
            self.shape = self.dims
            self.rmask = tuple(slice(None) if dim > 1 else 0 for dim in self.dims)
        else:
            self.shape = tuple(reversed(self.dims))
            self.rmask = tuple(slice(None) if dim > 1 else 0 for dim in self.data_shape)

        self.size = np.product(self.dims)

    @cached_property
    def mesh_center(self):
        """
        2D cell center mesh in native coordinates
        Returns:
        X, Y with shape for each: (dim[1],dim[0])
        """
        if len(self.rdims) == 1:
            return self.xn[self.rdims_ind[0]]
        elif len(self.rdims) == 2:
            X, Y = np.meshgrid(self.xn[self.rdims_ind[0]], self.xn[self.rdims_ind[1]])
            return self.T(X), self.T(Y)
        else:
            raise NotImplementedError("3D mesh not implemented yet")

    @cached_property
    def mesh_edge(self):
        """
        2D cell edge mesh in native coordinates
        Returns:
        X, Y with shape for each: (dim[1]+1,dim[0]+1)
        """
        if len(self.rdims) == 1:
            return self.xn[self.rdims_ind[0]]
        elif len(self.rdims) == 2:
            X, Y = np.meshgrid(self.xni[self.rdims_ind[0]], self.xni[self.rdims_ind[1]])
            return self.T(X), self.T(Y)
        else:
            raise NotImplementedError("3D mesh not implemented yet")

    @cached_property
    def mesh_center_cartesian(self):
        """TODO"""
        if len(self.rdims) != 2:
            raise NotImplementedError(
                "Projection to cartesian grid only implemented in 2D"
            )
        return transform_mesh(self, *self.mesh_center)

    @cached_property
    def mesh_edge_cartesian(self):
        """TODO"""
        if len(self.rdims) != 2:
            raise NotImplementedError(
                "Projection to cartesian grid only implemented in 2D"
            )
        return transform_mesh(self, *self.mesh_edge)

    def __getattr__(self, name: str):
        if name.startswith("_"):
            raise AttributeError(f"{type(self).__name__} has no attribute '{name}'")
        try:
            return getattr(self, self.mapping_grid[name])
        except KeyError:
            pass
        raise AttributeError(f"{type(self).__name__} has no attribute '{name}'")

    def __str__(self) -> str:
        mapping_inv = {value: key for key, value in self.mapping_grid.items()}
        return (
            f"PLUTO Grid Dimensions: {self.dims}, {self.coordinates} coordinate system\n"
            f"{mapping_inv['x1']}: {self.x1i[0]:.2f}..{self.x1i[-1]:.2f}, Lx1={self.Lx1:.2f}, N1={self.dims[0]}\n"
            f"{mapping_inv['x2']}: {self.x2i[0]:.2f}..{self.x2i[-1]:.2f}, Lx1={self.Lx2:.2f}, N1={self.dims[1]}\n"
            f"{mapping_inv['x3']}: {self.x3i[0]:.2f}..{self.x3i[-1]:.2f}, Lx1={self.Lx3:.2f}, N1={self.dims[2]}\n"
        )

    def __repr__(self) -> str:
        return f'{type(self).__name__}("{self.gridfile_path}", "{self.coordinates}")'

    def _repr_markdown_(self) -> str:
        """Jupyter pretty print"""
        return (
            f"**PLUTO Grid** Dimensions {self.dims}, {self.coordinates} coordinate system\n\n"
            "|   |   |   | L | N |\n"
            "|---|---|---|---|---|\n"
            f"|${self.mapping_tex['x1']}$|{self.x1i[0]:.2f}|{self.x1i[-1]:.2f}|{self.Lx1:.2f}|{self.dims[0]}|\n"
            f"|${self.mapping_tex['x2']}$|{self.x2i[0]:.2f}|{self.x2i[-1]:.2f}|{self.Lx2:.2f}|{self.dims[1]}|\n"
            f"|${self.mapping_tex['x3']}$|{self.x3i[0]:.2f}|{self.x3i[-1]:.2f}|{self.Lx3:.2f}|{self.dims[2]}|\n"
        )

    def __dir__(self):
        return object.__dir__(self) + list(self.mapping_grid.keys())


class GridSlice(Grid):
    def __init__(self, grid: Grid, slice_):
        self.slice = normalize_slice(slice_, grid.shape)
        self.slicer = None

        self.gridfile_path = None
        self.set_coordinate_system(grid.coordinates)

        self.T = grid.T
        self.indexing = grid.indexing

        # reverse slice depending on indexing
        slice_ijk = reversed(self.slice) if self.indexing == "kji" else self.slice

        self.xn = tuple(x[sl] for sl, x in zip(slice_ijk, grid.xn))
        self.xni = tuple(
            x[sl.start : sl.stop + sl.step : sl.step]
            for sl, x in zip(slice_ijk, grid.xni)
        )
        self.dxn = tuple(x[1:] - x[:-1] for x in self.xni)
        self.L = tuple(x[-1] - x[0] for x in self.xni)

        # reference in named attributes
        for i in range(3):
            setattr(self, f"x{i+1}", self.xn[i])
            setattr(self, f"x{i+1}i", self.xni[i])
            setattr(self, f"dx{i+1}", self.dxn[i])
            setattr(self, f"Lx{i+1}", self.L[i])

        self.dims = tuple(len(x) for x in self.xn)
        self.rdims = tuple(dim for dim in self.dims if dim > 1)
        self.rdims_ind = tuple(i for i, dim in enumerate(self.dims) if dim > 1)

        self.data_shape = None
        if self.indexing == "ijk":
            self.rmask = tuple(slice(None) if dim > 1 else 0 for dim in self.dims)
            self.shape = self.dims
        else:
            self.rmask = tuple(
                slice(None) if dim > 1 else 0 for dim in reversed(self.dims)
            )
            self.shape = tuple(reversed(self.dims))

        self.size = grid.size

        # TODO repr and str


def normalize_slice(slice_: tuple, shape: tuple) -> tuple:
    """Check bounds of 3D slice, and preserve 3d structure of array
    for 1-high direction slice

    Args:
        slice_ (tuple): 3D slice, consisting of `int` and `slice`
        shape (tuple): dimensions of domain

    Returns:
        tuple

    Raises:
        IndexError: if any of the bounds check don't work
    """
    if len(slice_) != 3:
        raise IndexError("Please specify 3D slice for clarity")

    newslice = []
    for dimslice, dimsize in zip(slice_, shape):
        if isinstance(dimslice, slice):
            start, stop, step = dimslice.start, dimslice.stop, dimslice.step

            if start is None:
                start = 0
            elif start >= dimsize or start < -dimsize:
                raise IndexError("Slice out of bounds")
            elif start < 0:
                start = dimsize + dimslice.start

            if stop is None:
                stop = dimsize
            elif stop > dimsize or stop < -dimsize:
                raise IndexError("Slice out of bounds")
            elif stop < 0:
                stop = dimsize + stop

            if step is None:
                step = 1

            newslice.append(slice(start, stop, step))

        else:
            start = dimslice
            if start < 0:
                start = dimsize + dimslice
            if not (0 <= start < dimsize):
                raise IndexError("Slice out of bounds")

            newslice.append(slice(start, start + 1, 1))

    return tuple(newslice)



"""Collection of entrypoints"""
import argparse


def format_ini():
    """Entrypoint to format `pluto.ini` file"""
    from .metadata import Pluto_ini

    parser = argparse.ArgumentParser(description="Fix indentation in pluto.ini")
    parser.add_argument("filename")
    args = parser.parse_args()

    Pluto_ini(args.filename).write()



"""PlutoData: Class to contain a single PLUTO output step"""
import numpy as np

try:
    import h5py
except ImportError:
    pass

from .grid import Grid, GridSlice
from .metadata import SimulationMetadata
from .misc import Slicer
from .plotting import plot


class PlutoData:
    """Object representing single PLUTO output step

    Attributes:
        n (int): Number of output in simulation
        t (float): Simulation time at output
        sim_dt (float): Simulation timestep at output
        nstep (int): Simulation steps at output
        metadata (plutoplot.io.SimulationMetadata): Metadata read from .out file
        grid (plutoplot.grid.Grid): Grid read from grid.out
        simulation (plutoplot.simulation.Simulation): Simulation this data step is part of.
            Can be None, as this doesn't have to be initialized from Simulation.
    """

    def __init__(
        self,
        n: int,
        metadata: SimulationMetadata,
        grid: Grid,
        simulation=None,
    ):
        self.n = n
        self.metadata = metadata
        self.grid = grid
        self.simulation = simulation

        self.t, self.sim_dt, self.nstep = (
            metadata.t[n],
            metadata.sim_dt[n],
            metadata.nstep[n],
        )

        self._data = {}
        self.slicer = Slicer(
            lambda slice_: PlutoDataSlice(self, self.grid.slicer[slice_])
        )

    def __getattr__(self, attr: str):
        """Get data / grid / metadata attributes

        This function only gets called if the PlutoData itself has no attribute with this name.

        Args:
            attr (str): name of attribute

        Returns:
            Any: requested attribute

        Raises:
            AttributeError: if `attr` is not an attribute of either data, grid, or metadata.
        """
        # don't forward private methods
        if attr.startswith("_"):
            raise AttributeError(f"{type(self).__name__} has no attribute '{attr}'")

        try:  # data variables
            return self[attr]
        except KeyError:
            pass
        try:  # grid
            return getattr(self.grid, attr)
        except AttributeError:
            pass
        try:  # simulation
            return self.simulation.__getattribute__(attr)
        except AttributeError:
            pass

        raise AttributeError(f"{type(self).__name__} has no attribute '{attr}'")

    def __delattr__(self, var):
        """Remove data array from memory"""
        try:
            del self[var]
        except KeyError as e:
            raise AttributeError(str(e))

    def __getitem__(self, var: str) -> np.memmap:
        """Get data array for variable

        Returns:
            :obj:`numpy.ndarray` or :obj:`numpy.memmap`
        """
        var_generic = self.grid.mapping_vars.get(var, var)
        try:
            return self._data[var_generic]
        except KeyError:
            if var_generic in self.metadata.vars:
                self._data[var_generic] = self._load_var(var_generic)
                return self._data[var_generic]
        raise KeyError(f"{type(self).__name__}: '{var}' is not a data variable")

    def __delitem__(self, var: str):
        """Delete data arrays to free memory

        Raises:
            KeyError
        """
        var_generic = self.grid.mapping_vars.get(var, var)
        try:
            del self._data[var_generic]
        except KeyError:
            if var_generic not in self.vars:
                raise KeyError(
                    f"{type(self).__name__}: '{var}' is not a data variable"
                ) from None

    def _load_var(self, varname) -> np.memmap:
        """Create memorymap to data

        Args:
            varname (str): variable name

        Returns:
            numpy.memmap: Memorymap to data
        """
        if self.metadata.format in ("dbl.h5", "flt.h5"):
            if self.grid.indexing == "ijk":
                raise NotImplementedError(
                    "ijk indexing not implemented for HDF5 outputs. Use indexing=kji"
                )
            try:
                self.h5file
            except AttributeError:
                try:
                    self.h5file = h5py.File(
                        self.metadata.data_path
                        / f"data.{self.n:04d}.{self.metadata.format}",
                        "r",
                    )
                except NameError:
                    raise ImportError(
                        "plutoplot: Optional dependency 'h5py' not installed, required for reading HDF5 files"
                    ) from None
            finally:
                return self._post_load_process(
                    varname, self.h5file[f"Timestep_{self.n}/vars/{varname}"]
                )

        if self.metadata.format in ("dbl", "flt"):
            if self.metadata.file_mode == "single":
                filename = f"data.{self.n:04d}.{self.metadata.format}"
                # byte offset of variable in binary file
                offset = (
                    self.metadata.charsize
                    * self.grid.size
                    * self.metadata.vars.index(varname)
                )
            else:
                filename = f"{varname}.{self.n:04d}.{self.metadata.format}"
                offset = 0
        elif self.format == "vtk":
            offset = self.metadata.vtk_offsets[varname]
            if self.file_mode == "single":
                filename = "data.{self.n:04d}.vtk"
            else:
                filename = "{var}.{self.n:04d}.vtk"

        return self._post_load_process(
            varname,
            self.grid.T(
                np.memmap(
                    self.metadata.data_path / filename,
                    dtype=self.metadata.binformat,
                    mode="c",
                    offset=offset,
                    shape=self.grid.data_shape,
                )
            ),
        )

    def _post_load_process(self, varname, data: np.ndarray):
        """Process data after loading from disk

        User method to process data after loading it from disk.
        Does nothing by default, can be overwritten in subclass.

        Args:
            varname
        """
        return data

    def plot(
        self, var: str, grid=None, label: str = None, show_title: bool = True, **kwargs
    ):
        """TODO"""
        if isinstance(var, str):
            varname = var
            var = getattr(self, var)
        else:
            varname = None
        if grid is None:
            grid = self.grid

        if label is None:
            label = "${}$".format(self.grid.mapping_tex.get(varname, varname))

        fig, ax = plot(var, grid, label=label, **kwargs)

        if show_title:
            ax.set_title(f"$t={self.t:.2f}$")

        return fig, ax

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}, output nr: {self.n}, time: {self.t}, simulation step: {self.nstep}\n"
            f"data directory: '{self.metadata.data_path}'\n"
            f"resolution: {self.grid.dims}, {self.grid.coordinates} coordinates\n"
            f"Variables: {' '.join(self.metadata.vars)}"
        )

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.n}, metadata={repr(self.metadata)}, "
            f"grid={repr(self.grid)}, simulation={repr(self.simulation)})"
        )

    def _repr_markdown_(self) -> str:
        """Jupyter pretty print"""
        return (
            f"**{type(self).__name__}**, output nr: {self.n}, time: {self.t}, simulation step: {self.nstep}  \n"
            f"data directory: `{self.metadata.data_path}`  \n"
            f"Data vars: `{'` `'.join(self.metadata.vars)}`  \n"
        ) + self.grid._repr_markdown_()

    def __dir__(self) -> list:
        return (
            object.__dir__(self)
            + self.metadata.vars
            + [attr for attr in dir(self.grid) if not attr.startswith("_")]
            + [attr for attr in dir(self.metadata) if not attr.startswith("_")]
            + list(self.grid.mapping_vars)
        )


class PlutoDataSlice(PlutoData):
    def __init__(self, parent: PlutoData, sliced_grid: Grid = None):
        self.parent = parent
        self.n = parent.n
        self.metadata = parent.metadata
        self.simulation = parent.simulation

        self.t, self.sim_dt, self.nstep = (
            parent.metadata.t[self.n],
            parent.metadata.sim_dt[self.n],
            parent.metadata.nstep[self.n],
        )

        self.grid = sliced_grid

        self.slicer = None

    def __getitem__(self, var: str) -> np.memmap:
        return self.parent[var][self.grid.slice]

    def __delitem__(self, var: str):
        del self.parent[var]

    def _load_var(self, varname):
        raise NotImplementedError("_load_var is only implemented on parent object")



"""Classes for PLUTO simulation Metadata and parsing of input files.

- :obj:`SimulationMetadata` reads PLUTO `*.out` files
- :obj:`PlutoIni` reads `pluto.ini` initialization files
- :obj:`Definitions_h` reads `definitions.h` compile time PLUTO configuration

"""
from collections import OrderedDict
from itertools import zip_longest
from pathlib import Path
from typing import Dict, List

import numpy as np


class SimulationMetadata:
    """Pluto simulation metadata reader and parser

    Read metadata from `format.out` files (e.g. `dbl.out`, `flt.out`).

    Attributes:
        path (Path): path to outputfile
        file_mode (str): PLUTO output in `"single"` or `"multiple"` files.
        vtk_offsets (:obj:`dict` of :obj:`int`): Byte offsets of data in VTK file
        t (:obj:`numpy.ndarray` of `numpy.float64`): simulation time of outputs
        dt (:obj:`numpy.ndarray` of `numpy.float64`): simulation time difference of outputs
        sim_dt (:obj:`numpy.ndarray` of `numpy.float64`): simulation timestep at outputs
        nstep (:obj:`numpy.ndarray` of `numpy.float64`): simulation step at outputs
        file_mode (str): `"single"` or `"multiple"`, PLUTO output mode for format
        vars (:obj:`list` of :obj:`str`): Variables available in output
        charsize (int): size of floating point number in bytes (4 or 8)
        binformat (str): binary format descriptor
    """

    def __init__(self, path: Path, format: str):
        """Read PLUTO simulation metadata from file

        Args:
            path (Path): path to outputfile or data directory. If this is a directory
                path, then the file at `{path}/{format}.out` will be read.
            format (Path): output format, e.g. `dbl`, `flt`, `vtk`
            length (int): number of outputs

        """
        self.path = Path(path)
        self.format = format
        if self.path.is_dir():
            self.path = self.path / f"{format}.out"
        self.data_path = self.path.parent

        self.read_vars(self.path, format)

        # read VTK offsets in file
        if format == "vtk":
            if self.file_mode == "single":
                self.vtk_offsets = vtk_offsets(self.path.parent / "data.0000.vtk")
            else:
                self.vtk_offsets = {}
                for var in self.vars:
                    self.vtk_offsets.update(
                        vtk_offsets(self.path.parent / f"{var}.0000.vtk")
                    )

    def read_vars(self, path: Path, format: str) -> None:
        """Read simulation step data and written variables

        Attributes set in this method are described in class docstring.

        Args:
            path (path): path to output file (`format.out`)
            format (str): PLUTO output format
        """
        with path.open() as f:
            lines = f.readlines()
            self.length = len(lines)
            # prepare arrays
            self.t = np.empty(self.length, float)
            self.sim_dt = np.empty(self.length, float)
            self.nstep = np.empty(self.length, int)

            # this information should be the same for all outputs
            file_mode, endianness, *self.vars = lines[0].split()[4:]
            self.file_mode = "single" if file_mode == "single_file" else "multiple"
            # binary format
            self.charsize = 8 if format == "dbl" else 4
            endianness = "<" if endianness == "little" else ">"
            if format == "vtk":
                endianness = ">"  # VTK has always big endian
            self.binformat = "{}f{}".format(endianness, self.charsize)

            # metadata for single timesteps
            for i, line in enumerate(lines):
                self.t[i], self.sim_dt[i], self.nstep[i] = line.split()[1:4]
            self.dt = self.t[1:] - self.t[:-1]

    def __repr__(self):
        return f"{type(self).__name__}('{self.path}','{self.format}')"


def vtk_offsets(path: Path) -> Dict[str, int]:
    """Read positions of vars in VTK legacy file

    Legacy VTK files contain header und binary data. This function
    extracts the byte offsets for the variables from the header

    Args:
        path (Path): path to VTK file

    Returns:
        :obj:`dict` of :obj:`str`: :obj:`int`: Byte offsets for variable data
    """
    offsets = {}
    with path.open("rb") as f:
        for l in f:
            if not l or l == b"\n":
                continue

            split = l.split()

            # skip coordinates (read in via gridfile)
            if split[0] in [i + b"_COORDINATES" for i in [b"X", b"Y", b"Z"]]:
                f.seek(int(split[1]) * 4 + 1, 1)

            if split[0] == b"CELL_DATA":
                bytesize = int(split[1]) * 4

            # save position of variables
            if split[0] == b"SCALARS":
                var = split[1].decode()
                f.readline()  # skip "LOOKUP_TABLE"
                offsets[var] = f.tell()
                f.seek(bytesize, 1)

    return offsets


class Pluto_ini(OrderedDict):
    """PLUTO runtime initialization parameters pluto.ini

    Parses and writes `pluto.ini` files.
    Access with dict-like interfaces:
    - `ini["section"]["name"]`
    - `ini["section", "name"]`
    - `ini["section/name"]`

    Example:
        >>> ini = Pluto_ini(path)
        >>> ini["Time"]["tstop"]

    Attributes:

    """

    sections = OrderedDict.values

    class Section(OrderedDict):
        """Pluto_ini Section

        Thin wrapper around :obj:`OrderedDict`, with some convenience functions.

        Attributes;
            name (str): name of section
        """

        def __init__(self, name, *args, **kwargs):
            """Create Pluto_ini section

            Args:
                name (str): name of section
                *args, **kwargs: passed to :obj:`OrderedDict` constructor
            """
            super().__init__(*args, **kwargs)
            self.name = name

        def _align(self) -> List[int]:
            """Find column widths for aligned section"""
            length = []
            for key, value in self.items():
                if isinstance(value, str):
                    length.append((len(key), len(value)))
                else:
                    length.append((len(key), *[len(v) for v in value]))
            return [max(i) for i in zip_longest(*length, fillvalue=0)]

        def __str__(self) -> str:
            """Output Section in pluto.ini format with aligned columns"""
            colwidth = self._align()
            out = f"[{self.name}]\n\n"
            for key, values in self.items():
                out += f"{key:<{colwidth[0]}}"
                values = values if isinstance(values, list) else [values]
                for width, value in zip(colwidth[1:], values):
                    out += f"  {value:>{width}}"
                out += "\n"
            return out

        def _repr_html_(self) -> str:
            """Pretty printing in Jupyter"""
            return "<table>" + self._repr_html_inner() + "</table>"

        def _repr_html_inner(self) -> str:
            """Helper function for _repr_html_()"""
            out = f'<thead><tr><th colspan=2 style="text-align: left">{self.name}</th></tr></thead><tbody>'
            for key, value in self.items():
                if isinstance(value, list):
                    value = "&nbsp;&nbsp;".join(value)
                out += f"<tr><td>{key}</td><td>{value}</td></tr>"
            out += "</tbody>"
            return out

    def __init__(self, path: Path):
        """Load pluto.ini from file

        Args:
            path (Path): path to pluto.ini file
        """
        super().__init__()
        self.path = Path(path)

        self.parse()

    def parse(self, txt: str = None) -> None:
        """Parse pluto.ini file

        Args:
            txt (:obj:, optional): parse string instead of file
        """
        if txt is None:
            with self.path.open() as f:
                lines = [l.strip() for l in f.readlines()]
        else:
            lines = [l.strip() for l in txt.split("\n")]

        section = None
        for line in lines:
            if not line:
                continue
            elif line[0] == "[" and line[-1] == "]":
                section = line[1:-1]
                self[section] = self.Section(section)
            else:
                segments = line.split()
                if len(segments) > 2:
                    self[section][segments[0]] = segments[1:]
                else:
                    self[section][segments[0]] = segments[1]

    def __repr__(self) -> str:
        return f"{type(self).__name__}('{self.path}')"

    def _repr_html_(self) -> str:
        """Pretty printing in Jupyter"""
        return (
            f"<table>"
            + "".join(section._repr_html_inner() for section in self.sections())
            + "</table>"
        )

    def __str__(self) -> str:
        """Convert to pluto.ini format, with columns aligned inside sections"""
        return "\n\n".join(str(section) for section in self.sections())

    def _split_key(self, key) -> tuple:
        # (section, key)
        if isinstance(key, tuple):
            return key
        sep = key.find("/")
        # "section/key"
        if sep > -1:
            return key[:sep], key[sep + 1 :]
        # "section"
        return (key,)

    def __getitem__(self, key):
        """ini[key] with multiple syntaxes:

        - `ini["section"]`
        - `ini["section","key"]
        - `ini["section/key"]
        """
        key = self._split_key(key)
        if len(key) == 2:
            return super().__getitem__(key[0])[key[1]]
        return super().__getitem__(key[0])

    def __setitem__(self, key, value):
        """ini[key] = value
        For `key` syntax check `__getitem__()`
        """
        key = self._split_key(key)
        if len(key) == 1:
            super().__setitem__(key[0], value)
        else:
            if key[0] in self:
                self[key[0]][key[1]] = value
            else:
                super().__setitem__(key[0], self.Section(key[0], {key[1]: value}))

    def write(self, path: Path = None) -> None:
        """Write pluto.ini to file.

        If no path is given, the read file is overwritten

        Args:
            path (:obj:`Path`, optional): Path to write to
        """
        if path is None:
            path = self.path

        with open(path, "w") as f:
            f.write(str(self))


class Definitions_h(OrderedDict):
    """PLUTO compile time definitions from definitions.h

    Todo:
        * Implement file writing (sections need to stay as is)
        * Better string output
    """

    def __init__(self, path: Path):
        """Load definitions.h from file

        Args:
            path (Path): path to definitions.h file
        """
        super().__init__()
        self.path = Path(path)
        self.parse()

    def parse(self, txt: str = None) -> None:
        """Parse defintions.h

        Args:
            txt (:obj:, optional): parse string instead of file
        """
        with self.path.open() as f:
            lines = [l.strip() for l in f.readlines()]

        for line in lines:
            if not line:
                continue
            segments = line.split()
            if segments[0] == "#define":
                self[segments[1]] = segments[2]

    def __repr__(self) -> str:
        return f"{type(self).__name__}('{self.path}')"

    def __str__(self) -> str:
        max_key = max(len(key) for key in self.keys())
        return "".join(f"{key:>{max_key}} = {value}\n" for key, value in self.items())

    def _repr_html_(self) -> str:
        """Pretty printing in Jupyter"""
        return (
            f"<table><tbody>"
            + "".join(
                f"<tr><td>{key}</td><td>{value}</td></tr>"
                for key, value in self.items()
            )
            + "</tbody></table>"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plutoplot']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3,<4.0', 'numpy>=1.19,<2.0']

extras_require = \
{'hdf5': ['h5py>=3.0,<4.0'], 'interactive': ['ipywidgets>=7.0,<8.0']}

entry_points = \
{'console_scripts': ['pluto-format-ini = plutoplot.scripts:format_ini',
                     'pluto-info = plutoplot.info:main']}

setup_kwargs = {
    'name': 'plutoplot',
    'version': '0.3.2',
    'description': 'Python package to load PLUTO simulation data',
    'long_description': '# plutoplot - Plutocode data loading and plotting library\n![GitHub release](https://img.shields.io/github/v/release/simske/plutoplot?include_prereleases&label=Github%20Release)\n![PyPI](https://img.shields.io/pypi/v/plutoplot)\n[![Documentation](https://readthedocs.org/projects/plutoplot/badge/?version=latest)](https://plutoplot.readthedocs.io/en/latest/?badge=latest)\n[![pre-commit](https://github.com/Simske/plutoplot/actions/workflows/pre-commit.yml/badge.svg)](https://github.com/Simske/plutoplot/actions/workflows/pre-commit.yml)\n[![testing](https://github.com/Simske/plutoplot/actions/workflows/tox.yml/badge.svg)](https://github.com/Simske/plutoplot/actions/workflows/tox.yml)\n![License](https://img.shields.io/github/license/simske/plutoplot)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n\n`plutoplot` is a Python package for loading, analyzing and plotting data from the [astrophysical gas dynamics simulation code PLUTO](http://plutocode.ph.unito.it/).\nBy handling simulations as a whole and loading data only when needed (lazy loading), it is fast and easy to use.\n\nIt features:\n- Loading on data actually needed (by lazy loading in memory mapping)\n- easy slicing into large datasets (e.g. plotting 2D slice of 3D dataset)\n- plotting non-cartesian coordinates projected into cartesian space\n- interactive plotting in Jupyter Notebooks\n- CLI tool to quickly acquire information about simulations\n- Access to parameters set in `definitions.h` and `pluto.ini`\n- Access to simulation metadata and grid properties\n- Indexing with `ijk` (like `pyPLUTO`) and `kji` (like PLUTO internally) schemes\n\nThere is currently no support for:\n- Particles\n- Adaptive grids (AMR)\n- 3D plotting (only plotting 2D slices of 3D data is supported.)\n\nTo get started with `plutoplot` check out the [documentation](https://plutoplot.readthedocs.io/en/latest/) and [examples](https://github.com/Simske/plutoplot/tree/main/test).\n\n## Installation\n`plutoplot` requires Python >=3.6. Make sure for installation to install in the right environment, and/or adapt the pip commands accordingly (e.g. `pip3`).\n\nThe base version of plutoplot can be installed from Pypi with:\n```bash\npip install --upgrade plutoplot\n```\n\nTo keep the installation size small the dependencies for interactive plotting and loading of HDF5 data are declared as optional dependencies.\nTo install the optional dependencies use either of the following, depeding on your needs:\n```bash\npip install --upgrade plutoplot[interactive]\npip install --upgrade plutoplot[hdf5]\npip install --upgrade plutoplot[interactive,hdf5]\n```\n\nTo install the current development version from this branch:\n```bash\npip install --upgrade "plutoplot @ git+https://github.com/simske/plutoplot"\n```\n\n## Quick Start\n### CLI: check simulation info\nTo quickly check the status of a PLUTO simulation, you can use the `pluto-info` command line tool:\n```bash\npluto-info /path/to/simulation\n```\nSample:\n```\n>> pluto-info test-problems/HD/Jet/01/\nPLUTO simulation at \'test-problems/HD/Jet/01\'\nPolar grid with dimensions (160, 1, 640)\nDomain: x1: 0.00e+00 .. 1.00e+01 (Lx1 = 1.00e+01)\n        x2: 0.00e+00 .. 1.00e+00 (Lx2 = 1.00e+00)\n        x2: 0.00e+00 .. 4.00e+01 (Lx3 = 4.00e+01)\nAvailable variables: rho vx1 vx2 vx3 prs\nData files:\n    Format dbl: 16 files, last time 15.0, data timestep 1.00e+00\n    Format flt: 61 files, last time 15.0, data timestep 2.50e-01\n\n```\n(note: depending on your Python environment, the entrypoint `pluto-info` might not be installed correctly. Check the [documentation for more details](https://plutoplot.readthedocs.io/en/latest/cli-tools)).\n\n### Quick start in Python scripts / Jupyter notebooks\n```python\nimport plutoplot as pp\n```\n#### Basics\nSimulations can be loaded by just providing the path to the simulation directory (directory with `definitions.h` and `pluto.ini`).\nThe data directory is the found automatically, as well as the grid geometry.\nThe formats are searched in the order (dbl, flt, vtk, dbl.h5, flt.h5) and the first found loaded, this can be overriden.\n```python\nsim = pp.Simulation("/path/to/simulation", format="flt")\n```\nThe simulation object provides access to all outputs (of this format), the grid and simulation metadata.\nA few examples of accesible attributes:\n```python\nsim.dims                  # domain resolution\nsim.x1, sim.x2, sim.x3    # cell centered coordinates, redirect to sim.grid.x{1,2,3}\nsim.x1i, sim.x2i, sim.x3i # cell interface coordinates\nsim.Lx1, sim.Lx2, sim.Lx3\n\nsim.t                     # simulation output times\nsim.vars                  # simulation output variables\n```\n\nThe data at the specific output steps can be accessed with the `sim[n]` syntax.\nThis results in a `PlutoData` object, which can be used to access the data:\n```python\nsim[-1].rho\n```\nThe data object as well as the data array are generated when needed, and the data arrays are memory mapped, which means only data actually accessed is read from disk.\nThe names of the data variables can be found with `sim.vars`.\n\nBe aware that data arrays are always 3-dimensional, even if the domain has a resolution of 1 in some directions.\nThis is for clarity and consistency, so that an axis always has a direct correspondence to its coordinates in the grid.\n\nSimulations are iterable:\n```python\nfor step in sim:\n    print("Total mean squared velocity at t={step.t}: ", (step.vx1**2 + step.vx2**2 + step.vx3**2).mean())\n```\nAnd also provide a `reduce()` function to run over the whole simulation:\n```python\nprs_mean = sim.reduce(lambda step: step.prs.mean())\n```\nThis results in an array containing the output of the supplied function for each step, in the example the mean pressure.\n\n#### Working with simulation slices\nTo only look at a part of a simulation we can slice into it.\nThis can be particularly useful for analyzing and plotting 2D slices of 3D data, as full 3D plotting is not supported yet.\nTo slice into a simulation the usual slicing syntax on `Simulation.slicer`:\n```python\nsliced_sim = sim.slicer[:,0,32:64]\n```\nThis sliced simulation object can now be used as the normal simulation object.\nThe grid and data arrays will automatically return the data for this slice.\n(Note: for clarity a 3D slice is always required. The data arrays will still always be 3D, see above).\n\n#### Plotting\n`plutoplot` provides automatic plotting for 1D & 2D simulations.\nTo plot slices of 3D simulations, first create a simulation slice (section above), then just follow normally.\n\nThe PlutoData object provides a `plot()` function, which will automatically project the data onto a cartesian grid, and annotating the axes and colorbar correctly:\n```python\nsim[-1].plot(\'rho\')\n```\n![Density plot](docs/docs/img/quick_start_plot_projected.jpg)\n\nThe projection can also be turned off, an arbitrary options which will be passed to matplotlib can be given:\n```python\nsim[-1].plot(\'rho\', projection=False)\n```\n![Density plot](docs/docs/img/quick_start_plot_not-projected.jpg)\n\nIf a plot of a custom quantitity is needed, a array instead of a variable name can be given:\n```\nstep = sim[-1]\nstep.plot(step.prs / step.rho, label="T")\n```\n\nIn Jupyter Notebooks interactive plotting can be used to step through the simulation steps.\n`Simulation.iplot()` has the same interface as `PlutoData.plot()`, but it will show a slider to step through simulation steps:\n```\nsim.iplot("rho")\n```\n![Interactive plot](docs/docs/img/quick_start_iplot.gif)\n\n## Testing\nThe test suite for `plutoplot` can be found in `test/` and is run by Github actions using `tox` and `pytest`.\n\nThe test suite is currently minimal and has low code coverage. It will be expanded in the future to ensure correctness.\n\n## Contributing\nIf you have any questions, suggestions or find a bug, feel free to open an [issue](https://github.com/Simske/plutoplot/issues)\nor [Pull Request](https://github.com/Simske/plutoplot/pulls), I am happy to accept contributions to make plutoplot better.\n\n`plutoplot` uses [pre-commit](https://pre-commit.com/) to ensure a consistend code style, please use it when opening a Pull Request.\n\n## License\nGPLv3\n',
    'author': 'Simeon Doetsch',
    'author_email': 'science@simske.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Simske/plutoplot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import ScalarFormatter
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .grid import Grid, GridSlice


def plot(
    data: np.ndarray,
    grid: Grid,
    *,
    ax=None,
    label: str = None,
    figsize=None,
    cbar=True,
    projection: bool = True,
    **mpl_kwargs,
) -> None:
    """Simple colorplot for 2-dim data"""

    if isinstance(grid, GridSlice) and data.shape == grid.shape:
        data = data[grid.slice]
    elif data.shape == grid.shape:
        pass
    else:
        raise RuntimeError("Plotting: Grid shape not compatible with data")

    if len(grid.rdims) == 1:
        if ax is None:
            if figsize is None:
                figsize = (16, 10)
            _, ax = plt.subplots(figsize=figsize)

        ax.set_xlabel(f"${grid.mapping_tex[f'x{grid.rdims_ind[0]+1}']}$")
        ax.set_ylabel(label)
        ax.grid()

        ax.plot(grid.xn[grid.rdims_ind[0]], data[grid.rmask], label=label, **mpl_kwargs)

    elif len(grid.rdims) == 2:
        if projection:
            (xlabel, ylabel), (X, Y) = grid.mesh_edge_cartesian

        else:
            xlabel = f"{grid.mapping_tex[f'x{grid.rdims_ind[0]+1}']}"
            ylabel = f"{grid.mapping_tex[f'x{grid.rdims_ind[1]+1}']}"
            X, Y = grid.mesh_edge

        if ax is None:
            if figsize is None:
                # set a figsize depending on aspect ratio
                ratio = (X.max() - X.min()) / (Y.max() - Y.min())
                if ratio < 16 / 9:
                    y_size = 10
                    x_size = y_size / ratio * 1.1
                else:
                    x_size = 16
                    y_size = x_size * ratio / 1.1
                figsize = (x_size, y_size)
            _, ax = plt.subplots(figsize=figsize)

        ax.set_xlabel(f"${xlabel}$")
        ax.set_ylabel(f"${ylabel}$")

        im = ax.pcolormesh(X, Y, data[grid.rmask], **mpl_kwargs)
        ax.set_aspect(1)
        if cbar:
            formatter = ScalarFormatter()
            formatter.set_powerlimits((-2, 2))
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size=0.15, pad=0.05)
            plt.colorbar(im, label=label, format=formatter, cax=cax)

            # plt.colorbar(im, label=label, format=formatter, ax=ax, pad=0.05)

    elif len(grid.rdims) == 3:
        raise NotImplementedError("3D plotting not supported (yet)")

    return ax.figure, ax



import multiprocessing
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from .grid import Grid
from .metadata import Definitions_h, Pluto_ini, SimulationMetadata
from .misc import Slicer, cached_property
from .plutodata import PlutoData, PlutoDataSlice


class Simulation:
    """
    Container class for PLUTO (http://plutocode.ph.unito.it/) output.
    Reads the metadata of all files in working directory (wdir), and
    loads individual files when needed.
    Simulation is subscriptable and iterable.

    Attributes:
        path (pathlib.Path): Path to simulation directory
            (directory with `pluto.ini`, `definitions.h`, etc)
        data_path (pathlib.Path): path to data directory (with `*.out` and data files)
        format (str): simulation format
        metadata (plutoplot.io.SimulationMetadata): Simulation metadata
        grid (plutoplot.grid.Grid): Simulation grid
    """

    supported_formats = ("dbl", "flt", "vtk", "dbl.h5", "flt.h5")
    DataObject = PlutoData

    def __init__(
        self,
        path: Path = ".",
        format: str = None,
        coordinates: str = None,
        indexing: str = "ijk",
    ):
        """Create Simulation object from PLUTO output directory

        Args:
            path (Path or str, optional): Path to simulation directory / data directory
            format (str, optional): PLUTO output format to load.
                Currently supported: 'dbl', 'flt', 'vtk', 'dbl.h5', 'flt.h5'.
                By default uses the first {format}.out file found.
            coordinates (str, optional): name of coordinates system of simulation.
                Will be read from gridfile by default
                Supported: (')
            indexing (str, optional): Array index convention. Supports ("ijk", "kji").

        Raises:
            FileNotFoundError: if no metadata or grid files are found.
            NotImplementedError: if unsupported format is requested
        """
        self.parent = None
        self.path = Path(path)

        ## Find data directory ##
        if (self.path / "grid.out").exists():
            self.data_path = self.path
        elif (self.path / "data" / "grid.out").exists():
            self.data_path = self.path / "data"
        else:
            try:
                from_ini = self.path / self.ini["Static Grid Output"]["output_dir"]
                if (from_ini / "grid.out").exists():
                    self.data_path = from_ini
                else:
                    raise FileNotFoundError()
            except FileNotFoundError:
                raise FileNotFoundError(
                    "Data directory with gridfile not found"
                ) from None

        ## Find data format
        self.format = None
        if format is None:
            for format in self.supported_formats:
                if (self.data_path / f"{format}.out").exists():
                    self.format = format
                    break
            if self.format is None:
                raise FileNotFoundError(
                    f"No Metadata file for formats {self.supported_formats} found in {self.data_path}"
                )
        else:
            if format not in self.supported_formats:
                raise NotImplementedError(f"Format '{format}' not supported")
            if (self.data_path / f"{format}.out").exists():
                self.format = format
            else:
                raise FileNotFoundError(
                    f"Metadata file {self.data_path / f'{format}.out'} not found."
                )

        ## Read metadata ##
        self.metadata = SimulationMetadata(self.data_path, self.format)

        ## Read grid ##
        # coordinate system will be read from gridfile if `coordinates is None`
        self.grid = Grid(self.data_path / "grid.out", coordinates, indexing=indexing)

        # slicer
        self.slicer = Slicer(
            lambda slice_: SimulationSlice(self, self.grid.slicer[slice_])
        )

        # PlutoData cache
        self._data = {}

    @cached_property
    def ini(self) -> Pluto_ini:
        """Read access to PLUTO runtime initialization file 'pluto.ini'"""
        return Pluto_ini(self.path / "pluto.ini")

    @cached_property
    def definitions(self) -> Definitions_h:
        """Read access to PLUTO compile time 'definitions.h' file"""
        return Definitions_h(self.path / "definitions.h")

    def __getattr__(self, attr: str):
        """Resolve attributes to metadata/data/grid attributes"""
        # Don't resolve private attributes
        if attr.startswith("_"):
            raise AttributeError(f"{type(self).__name__} has no attribute '{attr}'")

        try:  # grid
            return getattr(self.grid, attr)
        except AttributeError:
            pass
        try:  # metadata
            return getattr(self.metadata, attr)
        except AttributeError:
            pass
        try:  # data from last simulation step
            return self[-1][attr]
        except KeyError:
            pass

        raise AttributeError(f"{type(self).__name__} has no attribute '{attr}'")

    def _index(self, key: int) -> int:
        """Checks if index is in range and implements negative indexing"""
        if not isinstance(key, (int, np.integer)):
            raise IndexError("Data index has to be an integer")
        elif key >= len(self):
            raise IndexError("Data index out of range")
        elif key < 0:
            key = self._index(len(self) + key)
            if key < 0:
                raise IndexError("Data index out of range")
        return key

    def __getitem__(self, key: int) -> DataObject:
        """Access individual data frames, return them as PlutoData

        If file is already loaded, object is returned, otherwise data is loaded and returned
        """
        return self.get(key)

    def get(self, key: int, keep: bool = True) -> DataObject:
        """Get PLUTO output

        Access individual data frames as PlutoData object. If `keep=True` the object is cached.

        Args:
            key (int): Output number
            keep (:obj:`bool`, optional): Keep output step in memory

        Returns:
            :obj:`Simulation.DataObject` (by default: :obj:`PlutoData`)
        """
        key = self._index(key)

        try:
            return self._data[key]
        except KeyError:
            # load data frame
            data = self._load_data(key)
            if keep:
                self._data[key] = data
            return data

    def __delitem__(self, key: int):
        """Delete data object to free memory"""
        key = self._index(key)
        try:
            del self._data[key]
        except KeyError:
            pass

    def _load_data(self, key: int) -> DataObject:
        """Load data frame"""
        return self.DataObject(
            self._index(key), metadata=self.metadata, grid=self.grid, simulation=self
        )

    def __iter__(self) -> "SimulationIterator":
        """Iterate over all data frames"""
        return self.iter()

    def iter(self, *range_, keep: bool = False) -> "SimulationIterator":
        """Iterate over simulation

        Range argument definition the same as `range()`.

        Attributes:
            simulation (plutoplot.simulation.Simulation): Simulation to iterate
            keep (bool): Whether to keep the PlutoData objects in memory
            start (int): Iteration start
            stop (int): Iteration stop (exclusive)
            step (int): Iteration step

        Returns:
            SimulationIterator
        """
        return SimulationIterator(self, *range_, keep=keep)

    def reduce(
        self,
        func,
        range=(),
        dtype=None,
    ):
        """Reduce all simulation steps with function.

        Shape and dtype are implied from return value of `func()`.

        Args:
            func (function): function which takes a PlutoData object and returns scalar or numpy array.
            dtype (numpy.dtype): forced data type for result array (if None func() implies dtype)
            range (tuple): range tuple for iterator.

        Returns:
            numpy.ndarray: reduced data array
        """
        # run on first element to get shape and dtype
        first = np.array(func(self[0]))
        if dtype is None:
            dtype = first.dtype
        #
        if first.shape == () or first.shape == (1,):
            return np.fromiter(
                (func(d) for d in self.iter(*range)),
                dtype=dtype,
                count=len(self.iter(*range)),
            )
        else:
            shape = (len(self.iter(*range)), *first.shape)
            res = np.empty(shape, dtype=dtype)
            for i, d in enumerate(self.iter(*range)):
                res[i] = func(d)
            return res

    def reduce_parallel(self, func, range=(), processes=None, dtype=None):
        """Reduce all simulation steps with function in parallel

        Shape and dtype are implied from return value of `func()`.

        Args:
            func (function): function which takes a PlutoData object and returns scalar or numpy array.
            dtype (numpy.dtype): forced data type for result array (if None func() implies dtype)
            range (tuple): range tuple for iterator.

        Returns:
            numpy.ndarray: reduced data array
        """
        first = np.array(func(self[0]))
        if dtype is None:
            dtype = first.dtype

        shape = (len(self.iter(*range)), *first.shape)
        res = np.empty(shape, dtype=dtype)
        with multiprocessing.Pool(processes) as p:
            for i, d in enumerate(p.imap(func, self.iter(*range))):
                res[i] = d
        return res

    def plot(self, *args, n: int = -1, **kwargs) -> None:
        """
        Plot last data file, or data file n. All other arguments forwarded to PlutoData.plot()
        """
        return self[n].plot(*args, **kwargs)

    def iplot(self, *args, plot_func=None, n: int = -1, **kwargs) -> None:
        """
        Plot simulation interactively. All other arguments forwarded to PlutoData.plot()
        No return, because it would interfere with interactive output in Jupyter Notebook
        """

        # Use widget to choose timestep if inside Jupyter Notebook
        try:
            get_ipython
            import ipywidgets as widgets

            if plot_func is None:

                def handler(i):
                    self[i].plot(*args, **kwargs)
                    plt.show()

            else:

                def handler(i):
                    plot_func(self[i])

            plot = widgets.interactive(
                handler,
                i=widgets.IntSlider(min=0, max=len(self) - 1, value=self._index(n)),
            )
            plot.children[0].description = "Step"
            plot.children[0].layout.width = "40%"
            display(plot)

        except NameError:
            raise RuntimeError(
                "Code has to be run in Jupyter Notebook for interactive plotting"
            )

    def __len__(self) -> int:
        return self.metadata.length

    def clear(self) -> None:
        """Clear loaded data frames"""
        self._data.clear()

    def __str__(self) -> str:
        return (
            f"PLUTO {type(self).__name__}: path: '{self.path}', "
            f"data directory '$sim_path/{self.data_path.relative_to(self.path)}'\n"
            f"Data vars: {', '.join(self.metadata.vars)}\n"
            f"Data files: Format `{self.format}`: {len(self)} files,"
            f" last time {self.t[-1]}, data timestep {self.dt.mean():.2e}  \n"
        ) + str(self.grid)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}('{self.path}', format='{self.metadata.format}', "
            f"coordinates='{self.grid.coordinates}')"
        )

    def _repr_markdown_(self) -> str:
        """Jupyter pretty print"""
        return (
            f"**PLUTO {type(self).__name__}** path: `{self.path}`, "
            f"data directory `$sim_path/{self.data_path.relative_to(self.path)}`  \n"
            f"Data vars: `{'` `'.join(self.metadata.vars)}`  \n"
            f"Data files: Format `{self.format}`: {len(self)} files,"
            f"last time {self.t[-1]}, data timestep {self.dt.mean():.2e}  \n"
        ) + self.grid._repr_markdown_()

    def __dir__(self) -> list:
        return (
            object.__dir__(self)
            + self.metadata.vars
            + [attr for attr in dir(self.grid) if not attr.startswith("_")]
            + [attr for attr in dir(self.metadata) if not attr.startswith("_")]
            + list(self.grid.mapping_vars)
        )


class SimulationIterator:
    """Iterator for Simulation

    Range definition the same as `range()`.

    Attributes:
        simulation (plutoplot.simulation.Simulation): Simulation to iterate
        keep (bool): Whether to keep the PlutoData objects in memory
        start (int): Iteration start
        stop (int): Iteration stop (exclusive)
        step (int): Iteration step

    Yields:
        PlutoData
    """

    def __init__(self, simulation: Simulation, *range_, keep: bool = False):
        """Create SimulationIterator

        Args:
            simulation (plutoplot.simulation.Simulation): Simulation to iterate
            *range_ (int): (), `stop` or `start, stop` or `start, stop, step`
            keep (bool): Whether to keep the PlutoData objects in memory
        """
        self.simulation = simulation
        self.keep = keep
        self.start, self.stop, self.step = 0, len(self.simulation), 1
        if len(range_) == 1:
            self.stop = self.simulation._index(range_[0])
        elif len(range_) == 2:
            self.start = self.simulation._index(range_[0])
            self.stop = self.simulation._index(range_[1])
        elif len(range_) == 3:
            self.start = self.simulation._index(range_[0])
            self.stop = self.simulation._index(range_[1])
            self.step = self.simulation._index(range_[2])
        elif len(range_) > 3:
            raise TypeError("Too many arguments for SimulationRange")

        self._iterator = iter(range(self.start, self.stop, self.step))

    def __len__(self):
        return len(range(self.start, self.stop, self.step))

    def __next__(self):
        return self.simulation.get(next(self._iterator), keep=self.keep)

    def __iter__(self):
        return self

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({repr(self.simulation)}, {self.start}, "
            f"{self.stop}, {self.step}, keep={self.keep})"
        )


class SimulationSlice(Simulation):
    def __init__(self, parent, sliced_grid):
        self.parent = parent
        self.path = self.parent.path
        self.data_path = self.parent.data_path
        self.format = self.parent.format
        self.metadata = self.parent.metadata

        self.grid = sliced_grid

        self.slicer = None

    def get(self, key: int, keep: bool = True):
        return PlutoDataSlice(self.parent.get(key, keep), sliced_grid=self.grid)
